#  _InvalidShapeError.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


class InvalidShapeError(RuntimeError):
    def __init__(self, message: str, *matrix_shapes: str):
        if matrix_shapes is None or len(matrix_shapes) == 0:
            super().__init__('Invalid shape ' + message)
        else:
            super().__init__('Invalid shapes '
                             + ', '.join([m for m in matrix_shapes])
                             + ' '
                             + message
                             )
